==> fary-1.1.0.tgz: package/src/middlewares/proxy/index.ts <==
import log = require('fary-log');
import http = require('http');
import Koa = require('koa');
import koaHttpProxy = require('koa-better-http-proxy');
import ENV = require('../../utils/env');
import isDomain = require('../../utils/is-domain');
import serverConfig = require('../../utils/server-config');
import ProxyLoadBalancer from './proxy-load-balancer';

import { IProxyMiddlewareOptions, IServerConfig } from '../../types';

const { IS_DEV, IS_SERVER_PRODUCTION } = ENV;
const serverConfigPath: string = serverConfig.path;
const serverConfigValue: IServerConfig = serverConfig.value;

const logger = log('fary:proxy');

interface IProxy {
  [key: string]: string;
}

interface IProxyReqOpts {
  headers?: {
    [key: string]: string,
  };
}

interface ILoadBalancedProxy {
  [key: string]: ProxyLoadBalancer;
}

/**
 * 获取负载均衡对象配置
 * get load balanced proxy
 * @param {IProxy} proxy - original proxy config
 * @return {ILoadBalancedProxy} *
 */
const getLoadBalancedProxy = (proxy: IProxy) => {
  return Object.keys(proxy).reduce((result: ILoadBalancedProxy, key) => {
    result[key] = new ProxyLoadBalancer(proxy[key], serverConfigValue.proxyLoadBalanceOptions);
    return result;
  }, {});
};

let prodLoadBalancedProxy: ILoadBalancedProxy;
if (!IS_DEV) {
  prodLoadBalancedProxy = getLoadBalancedProxy(serverConfigValue.proxy);
}

/**
 * load proxy config
 * @return {*} proxy result
 */
function getProxyConfig() {
  if (!IS_DEV) {
    return prodLoadBalancedProxy;
  }
  // in development mode, clear the cache every time when invoked to improve developing experience!
  delete require.cache[serverConfigPath];
  const newestServerConfig = require(serverConfigPath);
  logger.log('Got newest proxy mapping:', newestServerConfig.proxy);

  return getLoadBalancedProxy(newestServerConfig.proxy);
}

/**
 * proxy middleware
 * @return {Function} koa middleware
 */
export = ({
  postProxy = (d: Buffer) => d,
  preProxy = (o: koaHttpProxy.IRequestOption) => o,
}: IProxyMiddlewareOptions = {}) => {
  return async function ProxyMiddleware(ctx: Koa.Context, next: () => Promise<any>) {
    const url: string = ctx.url;
    let proxyTarget: string = '';
    let proxyConfig: ILoadBalancedProxy = getProxyConfig();

    // only when non *real production environment*, allow the proxy debugging
    if (!IS_SERVER_PRODUCTION) {
      // if the `proxyconfig` header is present, then merge it to `proxyConfig` for debugging
      const headerProxyConfigStr: string = ctx.headers.proxyconfig;
      if (headerProxyConfigStr) {
        let headerProxyConfig: IProxy;
        logger.log(`'proxyconfig' detected in headers: '${headerProxyConfigStr}'`);
        try {
          headerProxyConfig = JSON.parse(headerProxyConfigStr);
        } catch (e) {
          logger.warn('proxyconfig parsing FAILED, please check it.');
          headerProxyConfig = {};
        }
        proxyConfig = {
          ...proxyConfig,
          ...getLoadBalancedProxy(headerProxyConfig),
        };
        logger.log('Merged proxy config is:', proxyConfig);
      }
    }
    // find the one which match the prefix of `url`
    for (const prefix of Object.keys(proxyConfig)) {
      if (url.startsWith(prefix)) {
        const loadBalancedTarget: ProxyLoadBalancer = proxyConfig[prefix];
        const target: string = loadBalancedTarget.pick();
        // avoid the multiple proxy problem when local development
        if (!IS_DEV || !isDomain(target)) {
          ctx.url = url.replace(prefix, '');
        }
        proxyTarget = target;
        logger.log(`Proxy Found: '${prefix}' => '${proxyTarget}'`);
        break;
      }
    }
    if (!proxyTarget) {
      logger.warn('Proxy NOT FOUND, skipped');
      return next();
    }
    logger.log(`Request '${url}' will be proxied to '${proxyTarget + ctx.url}'`);

    return koaHttpProxy(proxyTarget, {
      // do not parse body, do not limit body size
      parseReqBody: false,
      /**
       * callback before proxy request
       * @param {Object} proxyReqOpts - proxy request options
       * @param {ctx} ctx - koa ctx
       * @return {Promise.<*>} *
       */
      async proxyReqOptDecorator(proxyReqOpts: koaHttpProxy.IRequestOption, context: Koa.Context) {
        // remove header `referer` and `origin` to avoid CSRF interception problem by backend when development
        if (IS_DEV) {
          delete proxyReqOpts.headers.referer;
          delete proxyReqOpts.headers.origin;
        }
        // record the timestamp for proxy duration
        context.state.proxyStartTimestamp = Date.now();
        return Promise.resolve(proxyReqOpts)
          .then(preProxy);
      },
      /**
       * callback after proxy response
       * @param {Response} proxyRes - proxy response
       * @param {Object} proxyResData - response data
       * @param {ctx} ctx - koa ctx
       * @return {Promise.<*>} *
       */
      async userResDecorator(proxyRes: http.IncomingMessage, proxyResData: Buffer, context: Koa.Context) {
        logger.info('ProxyRes headers:', proxyRes.headers);
        const location = `${proxyTarget}${context.url}`;
        logger.log(
          // tslint:disable-next-line
          `Proxy request '${location}' completed(${proxyRes.statusCode}), costing ${Date.now() - context.state.proxyStartTimestamp}ms.`);
        return Promise.resolve(proxyResData)
          .then(postProxy);
      },
    })(ctx, next);
  };
};

==> fary-1.3.1.tgz: package/src/middlewares/proxy/index.ts <==
import log = require('fary-log');
import http = require('http');
import Koa = require('koa');
import koaHttpProxy = require('koa-better-http-proxy');
import ENV = require('../../utils/env');
import isDomain = require('../../utils/is-domain');
import serverConfig = require('../../utils/server-config');
import ProxyLoadBalancer from './proxy-load-balancer';

import { IProxyMiddlewareOptions, IServerConfig } from '../../types';

const { IS_DEV, IS_SERVER_PRODUCTION } = ENV;
const serverConfigPath: string = serverConfig.path;
const serverConfigValue: IServerConfig = serverConfig.value;

const logger = log('fary:proxy');

interface IProxy {
  [key: string]: string;
}

interface IProxyReqOpts {
  headers?: {
    [key: string]: string,
  };
}

interface ILoadBalancedProxy {
  [key: string]: ProxyLoadBalancer;
}

/**
 * 获取负载均衡对象配置
 * get load balanced proxy
 * @param {IProxy} proxy - original proxy config
 * @return {ILoadBalancedProxy} *
 */
const getLoadBalancedProxy = (proxy: IProxy) => {
  return Object.keys(proxy).reduce((result: ILoadBalancedProxy, key) => {
    result[key] = new ProxyLoadBalancer(proxy[key], serverConfigValue.proxyLoadBalanceOptions);
    return result;
  }, {});
};

let prodLoadBalancedProxy: ILoadBalancedProxy;
if (!IS_DEV) {
  prodLoadBalancedProxy = getLoadBalancedProxy(serverConfigValue.proxy);
}

/**
 * load proxy config
 * @return {*} proxy result
 */
function getProxyConfig() {
  if (!IS_DEV) {
    return prodLoadBalancedProxy;
  }
  // in development mode, clear the cache every time when invoked to improve developing experience!
  delete require.cache[serverConfigPath];
  const newestServerConfig = require(serverConfigPath);
  logger.log('Got newest proxy mapping:', newestServerConfig.proxy);

  return getLoadBalancedProxy(newestServerConfig.proxy);
}

/**
 * proxy middleware
 * @return {Function} koa middleware
 */
export = ({
  postProxy = (proxyRes: http.IncomingMessage, d: Buffer) => d,
  preProxy = (o: koaHttpProxy.IRequestOption, ctx: Koa.Context) => o,
}: IProxyMiddlewareOptions = {}) => {
  return async function ProxyMiddleware(ctx: Koa.Context, next: () => Promise<any>) {
    const url: string = ctx.url;
    let proxyTarget: string = '';
    let proxyConfig: ILoadBalancedProxy = getProxyConfig();

    // only when non *real production environment*, allow the proxy debugging
    if (!IS_SERVER_PRODUCTION) {
      // if the `proxyconfig` header is present, then merge it to `proxyConfig` for debugging
      const headerProxyConfigStr: string = ctx.headers.proxyconfig;
      if (headerProxyConfigStr) {
        let headerProxyConfig: IProxy;
        logger.log(`'proxyconfig' detected in headers: '${headerProxyConfigStr}'`);
        try {
          headerProxyConfig = JSON.parse(headerProxyConfigStr);
        } catch (e) {
          logger.warn('proxyconfig parsing FAILED, please check it.');
          headerProxyConfig = {};
        }
        proxyConfig = Object.assign({}, proxyConfig, getLoadBalancedProxy(headerProxyConfig));
        logger.log('Merged proxy config is:', proxyConfig);
      }
    }
    // find the one which match the prefix of `url`
    for (const prefix of Object.keys(proxyConfig)) {
      if (url.startsWith(prefix)) {
        const loadBalancedTarget: ProxyLoadBalancer = proxyConfig[prefix];
        const target: string = loadBalancedTarget.pick();
        // avoid the multiple proxy problem when local development
        if (!IS_DEV || !isDomain(target)) {
          ctx.url = url.replace(prefix, '');
        }
        proxyTarget = target;
        logger.log(`Proxy Found: '${prefix}' => '${proxyTarget}'`);
        break;
      }
    }
    if (!proxyTarget) {
      logger.warn('Proxy NOT FOUND, skipped');
      return next();
    }
    logger.log(`Request '${url}' will be proxied to '${proxyTarget + ctx.url}'`);

    return koaHttpProxy(proxyTarget, {
      // do not parse body, do not limit body size
      parseReqBody: false,
      /**
       * callback before proxy request
       * @param {Object} proxyReqOpts - proxy request options
       * @param {ctx} ctx - koa ctx
       * @return {Promise.<*>} *
       */
      async proxyReqOptDecorator(proxyReqOpts: koaHttpProxy.IRequestOption, context: Koa.Context) {
        // remove header `referer` and `origin` to avoid CSRF interception problem by backend when development
        if (IS_DEV) {
          delete proxyReqOpts.headers.referer;
          delete proxyReqOpts.headers.origin;
        }
        // record the timestamp for proxy duration
        context.state.proxyStartTimestamp = Date.now();
        return Promise.resolve(preProxy(proxyReqOpts, context));
      },
      /**
       * callback after proxy response
       * @param {Response} proxyRes - proxy response
       * @param {Object} proxyResData - response data
       * @param {ctx} ctx - koa ctx
       * @return {Promise.<*>} *
       */
      async userResDecorator(proxyRes: http.IncomingMessage, proxyResData: Buffer, context: Koa.Context) {
        logger.info('ProxyRes headers:', proxyRes.headers);
        const location = `${proxyTarget}${context.url}`;
        logger.log(
          // tslint:disable-next-line
          `Proxy request '${location}' completed(${proxyRes.statusCode}), costing ${Date.now() - context.state.proxyStartTimestamp}ms.`);
        return Promise.resolve(postProxy(proxyRes, proxyResData, context));
      },
    })(ctx, next);
  };
};
